package main

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"net/http"
	"os"
	"strings"
)

type oioExporterHandler struct {
	exporter *OioExporter

	hostLabel string
	prefix    string
}

func Listen(ctx context.Context, host, bind, prefix string, exporter *OioExporter) {
	handler := &oioExporterHandler{
		exporter:  exporter,
		hostLabel: fmt.Sprintf(`host="%s"`, host),
		prefix:    prefix,
	}

	serveMux := http.NewServeMux()
	serveMux.HandleFunc("/metrics", handler.handleExport)
	serveMux.HandleFunc("/config", handler.handleConfig)
	serveMux.HandleFunc("/reload", handler.handleReload)
	serveMux.HandleFunc("/daemons", handler.handleDaemons)
	server := &http.Server{Addr: bind, Handler: serveMux}

	go func() {
		log.Infof("Running API at http://%s", bind)
		err := server.ListenAndServe()
		if err != nil && !errors.Is(err, http.ErrServerClosed) {
			log.Fatalf("HTTP server error: %s", err)
		}
	}()

	<-ctx.Done()

	if err := server.Shutdown(context.TODO()); err != nil {
		log.Warnf("Error shuting down the server: %s", err)
	}
}

// Export prometheus metrics
func (h *oioExporterHandler) handleExport(w http.ResponseWriter, r *http.Request) {
	metrics := h.exporter.metrics
	filter := r.URL.Query().Get("filter")

	metrics.RLock()
	metrics.Clean()
	for _, metric := range metrics.Measures {
		if filter != "" && !strings.HasPrefix(h.prefix+metric.Name, filter) {
			continue
		}

		labelStr := []string{h.hostLabel}
		for _, labelPair := range metric.Labels {
			labelStr = append(labelStr, labelPair[0]+"=\""+labelPair[1]+"\"")
		}

		if metric.Accumulator != nil {
			for k, v := range metric.Accumulator.Values() {
				fmt.Fprintf(w, "%s%s{%s} %f\n", h.prefix, metric.Name+"_"+k, strings.Join(labelStr, ","), v)
			}
		} else {
			fmt.Fprintf(w, "%s%s{%s} %f\n", h.prefix, metric.Name, strings.Join(labelStr, ","), metric.Value)
		}
	}
	metrics.RUnlock()
}

// Reload local service configuration
func (h *oioExporterHandler) handleReload(w http.ResponseWriter, r *http.Request) {
	if err := h.exporter.ReloadLocal(); err != nil {
		if err == ERR_ALREADY_RELOADING {
			w.WriteHeader(http.StatusConflict)
		} else {
			w.WriteHeader(http.StatusInternalServerError)
		}

		fmt.Fprint(w, err)

		return
	}

	_, _ = w.Write([]byte("OK"))
}

// Serve local service configuration
func (h *oioExporterHandler) handleConfig(w http.ResponseWriter, r *http.Request) {
	file, err := os.Open(h.exporter.inventoryPath)
	if err != nil {
		log.Errorf("Failed to open config file: %s", err)
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprint(w, err)
		return
	}
	if _, err := io.Copy(w, file); err != nil {
		log.Errorf("Could not serve config: %s", err)
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprint(w, err)
		return
	}
}

// Dump the running daemons
func (h *oioExporterHandler) handleDaemons(w http.ResponseWriter, r *http.Request) {
	daemons := make(map[string]DaemonsState)
	for exporterID, inventoryDaemons := range h.exporter.daemons {
		daemons[exporterID] = inventoryDaemons.daemonsState
	}

	if err := json.NewEncoder(w).Encode(daemons); err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprint(w, err)
	}
}
