package clog

import (
	"bytes"
	"context"
	"encoding/base64"
	"fmt"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"text/template"
	"time"

	"github.com/hpcloud/tail"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

var log = _log.New("log")

// TODO add Timeout

var serviceRegex = regexp.MustCompile(`^([^-]+)-(.+)-(\d+(?:\.\d+)?)$`)

var tailConfig = tail.Config{
	Logger: tail.DiscardingLogger,
	Follow: true,
	// This prevents channel closure on file moves
	ReOpen: true,
	// This prevents tail from re-parsing previous lines
	Location: &tail.SeekInfo{Offset: 0, Whence: 2},
}

type logDaemonFactory struct {
	defaultPatterns string
	patternsPath    string

	patternsYaml     PatternsYaml
	patternsVersions map[string]int
}

func NewFactory(defaultPatterns, patternsPath string) daemon.DaemonFactoryI {
	return &logDaemonFactory{
		defaultPatterns:  defaultPatterns,
		patternsPath:     patternsPath,
		patternsVersions: make(map[string]int),
	}
}

func (df *logDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})

	namespace := inventory.Namespace

	for serviceType, services := range inventory.Services {
		logYamlService, ok := df.patternsYaml[serviceType]
		if !ok {
			continue
		}

		for _, service := range services {
			for logTpl := range logYamlService {
				serviceID := service.ID
				daemonHash := fmt.Sprintf("%s:%s", serviceID, logTpl)
				patternHash := fmt.Sprintf("%s:%s", serviceType, logTpl)
				params[daemonHash] = &logDaemonParams{
					Namespace:   namespace,
					ServiceID:   serviceID,
					ServiceIP:   service.IP,
					ServicePort: service.Port,
					ServiceType: serviceType,
					LogTpl:      logTpl,
					Version:     df.patternsVersions[patternHash],
				}
			}
		}
	}

	return params
}

func (df *logDaemonFactory) Reload() error {
	defaultPatterns := df.defaultPatterns
	patternsPath := df.patternsPath

	var patternsYaml PatternsYaml

	if patternsPath == "" {
		data, err := base64.StdEncoding.DecodeString(defaultPatterns)
		if err != nil {
			return fmt.Errorf("Failed to base64-decode default patterns: %w", err)
		}

		if err := yaml.Unmarshal(data, &patternsYaml); err != nil {
			return err
		}
		log.Info("Logs definition loaded from defaults")
	} else {
		f, err := os.Open(patternsPath)
		if err != nil {
			return err
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&patternsYaml); err != nil {
			return err
		}
		log.Infof("Logs definition loaded from %s", patternsPath)
	}

	for serviceType, logYamlService := range patternsYaml {
		for logTpl, logPatterns := range logYamlService {
			for _, logPattern := range logPatterns {
				patternType := logPattern.Type
				pattern := logPattern.Pattern

				switch patternType {
				case "access":
					if logPattern.RespTimeDivider == 0 {
						logPattern.RespTimeDivider = 1
					}
				case "match":
					if logPattern.Name == "" {
						logPattern.Name = pattern
					}
				default:
					return fmt.Errorf("Unknown pattern type '%s' for service '%s'", patternType, serviceType)
				}

				patternRe, err := regexp.Compile(pattern)
				if err != nil {
					return fmt.Errorf("Failed to compile pattern `%s` for service '%s': %w", pattern, serviceType, err)
				}

				logPattern.patternRe = patternRe
			}

			prevLogService, ok := df.patternsYaml[serviceType]
			if !ok {
				continue
			}

			prevLogPatterns, ok := prevLogService[logTpl]
			if !ok {
				continue
			}

			if !logPatterns.Equal(prevLogPatterns) {
				logPatternHash := fmt.Sprintf("%s:%s", serviceType, logTpl)
				log.Infof("Pattern '%s' updated", logPatternHash)

				df.patternsVersions[logPatternHash]++
			}
		}
	}

	df.patternsYaml = patternsYaml

	log.Infof("parsed log patterns definitions for %d services", len(patternsYaml))

	return nil
}

func (df *logDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*logDaemonParams)
	namespace := params.Namespace
	serviceType := params.ServiceType
	logTpl := params.LogTpl

	return &logDaemon{
		namespace:   namespace,
		serviceID:   params.ServiceID,
		serviceIP:   params.ServiceIP,
		servicePort: params.ServicePort,
		serviceType: serviceType,
		logTpl:      logTpl,
		logPatterns: df.patternsYaml[serviceType][logTpl],
		prefix:      path.Join("/var/log/oio/sds", namespace),
		timeToRetry: 5 * time.Second,
	}
}

type LogPattern struct {
	Type            string  `yaml:"type"`
	Name            string  `yaml:"name"`
	Pattern         string  `yaml:"pattern"`
	RespTimeDivider float64 `yaml:"resp_time_divider"`
	patternRe       *regexp.Regexp
}

type LogPatterns []*LogPattern

type PatternsYaml map[string]map[string]LogPatterns

func (logPatterns1 LogPatterns) Equal(logPatterns2 LogPatterns) bool {
	if len(logPatterns1) != len(logPatterns2) {
		return false
	}

	for i, lp1 := range logPatterns1 {
		lp2 := logPatterns2[i]
		if !lp1.Equal(lp2) {
			return false
		}
	}

	return true
}

func (lp1 *LogPattern) Equal(lp2 *LogPattern) bool {
	return lp2 != nil &&
		lp1.Type == lp2.Type &&
		lp1.Name == lp2.Name &&
		lp1.Pattern == lp2.Pattern &&
		lp1.RespTimeDivider == lp2.RespTimeDivider

}

type logDaemonParams struct {
	Namespace   string
	ServiceID   string
	ServiceIP   string
	ServicePort int
	ServiceType string
	LogTpl      string
	Version     int
}

type logDaemon struct {
	daemon.BaseDaemon

	namespace   string
	serviceID   string
	serviceIP   string
	servicePort int
	serviceType string
	logTpl      string
	logPatterns LogPatterns
	prefix      string
	timeToRetry time.Duration

	serviceAddr string
	logPath     string
}

// Execute only once
func (d *logDaemon) Interval() time.Duration {
	return time.Duration(9e18)
}

func (d *logDaemon) Delay() time.Duration {
	return 0
}

func (d *logDaemon) Init(ctx context.Context) error {
	serviceID := d.serviceID
	serviceType := d.serviceType
	logTpl := d.logTpl
	prefix := d.prefix

	matches := serviceRegex.FindStringSubmatch(serviceID)
	if matches == nil || len(matches) != 4 {
		return fmt.Errorf("Could not expand service name for service '%s'", serviceID)
	}

	serviceName := matches[2]
	serviceIndex := matches[3]
	data := map[string]string{
		"Name":  serviceName,
		"Index": serviceIndex,
	}

	// FIXME: compile template after parsing configuration file, not here
	tpl := template.Must(template.New(serviceID).Parse(logTpl))

	var buffer bytes.Buffer
	if err := tpl.Execute(&buffer, data); err != nil {
		return fmt.Errorf("Error while templating logpath '%s' for service '%s': %w", logTpl, serviceID, err)
	}

	d.logPath = path.Join(prefix, fmt.Sprintf("%s-%s", serviceType, serviceIndex), buffer.String())
	d.serviceAddr = fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)

	return nil
}

// Collect is blocking
func (d *logDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	logPath := d.logPath
	timeToRetry := d.timeToRetry
	doneC := ctx.Done()

	var t *tail.Tail
	for {
		var err error
		t, err = tail.TailFile(logPath, tailConfig)
		if err != nil {
			m.IncrError("log", d.serviceType)
			log.Warnf("could not watch file '%s': %s. Retrying in %s", logPath, err, timeToRetry)

			select {
			case <-doneC:
				return nil
			case <-time.After(timeToRetry):
			}
			continue
		}

		break
	}

	defer func() {
		if err := t.Stop(); err != nil {
			log.Warnf("error stopping tailing file '%s': %s", logPath, err)
		}
		t.Cleanup()
	}()

	var line *tail.Line
	for {
		select {
		case <-doneC:
			return nil
		default:
		}

		select {
		case <-doneC:
			return nil
		case line = <-t.Lines:
		}

		// If the file does not exist yet
		// The channel outputs nils for some reason
		if line == nil {
			select {
			case <-doneC:
				return nil
			case <-time.After(timeToRetry):
			}
			continue
		}

		if line.Err != nil {
			log.Warnf("error getting line: %s", line.Err)
			continue
		}

		for _, logPattern := range d.logPatterns {
			if err := d.handleLine(line.Text, logPattern, m); err != nil {
				m.IncrError("log", d.serviceType)
				log.Warnf("error handling line from file '%s'(%s): %s", logPath, logPattern.Type, err)
			}
		}
	}
}

func (d *logDaemon) handleLine(line string, logPattern *LogPattern, m *common.MetricData) error {
	labels := [][]string{
		[]string{"service_addr", d.serviceAddr},
		[]string{"service", d.serviceID},
		[]string{"service_type", d.serviceType},
		[]string{"namespace", d.namespace},
	}

	switch logPattern.Type {
	case "access":
		return d.handleLineAccess(line, logPattern, labels, m)
	case "match":
		return d.handleLineMatch(line, logPattern, labels, m)
	}

	return nil
}

func (d *logDaemon) handleLineMatch(line string, logPattern *LogPattern, labels [][]string, m *common.MetricData) error {
	if !logPattern.patternRe.MatchString(line) {
		return nil
	}
	m.Acc(&common.Metric{
		Name:   "log_matches",
		Labels: append(labels, []string{"match", logPattern.Name}),
		Value:  1,
		Type:   "counter",
	})

	return nil
}

func (d *logDaemon) handleLineAccess(line string, logPattern *LogPattern, baseLabels [][]string, m *common.MetricData) error {
	matches := logPattern.patternRe.FindStringSubmatch(line)
	names := logPattern.patternRe.SubexpNames()[1:]
	if len(matches) != len(names)+1 {
		// TODO: add a debug line to see when a line is skipped
		return nil
	}

	if common.Index(names, "method") == -1 || common.Index(names, "code") == -1 {
		return nil
	}

	baseLabels = append(baseLabels, []string{"method", matches[1+common.Index(names, "method")]})
	baseLabels = append(baseLabels, []string{"code", matches[1+common.Index(names, "code")]})
	for idx, tag := range names {
		if strings.HasPrefix(tag, "tag_") {
			baseLabels = append(baseLabels, []string{strings.Split(tag, "tag_")[1], matches[idx+1]})
		}
	}

	for idx, name := range names {
		if common.Index([]string{"bytes_in", "bytes_out", "resp_time", "code"}, name) == -1 {
			continue
		}

		value := matches[idx+1]
		if value == "" || value == "-" {
			continue
		}

		vFloat, err := strconv.ParseFloat(value, 4)
		if err != nil {
			return fmt.Errorf("Could not convert '%s' to float: %w", name, err)
		}

		labels := make([][]string, len(baseLabels), len(baseLabels)+1)
		copy(labels, baseLabels)

		switch name {
		case "bytes_in":
			m.Acc(&common.Metric{
				Name:   "log_size_bytes",
				Labels: append(labels, []string{"direction", "in"}),
				Value:  vFloat,
				Type:   "stats",
			})
		case "bytes_out":
			m.Acc(&common.Metric{
				Name:   "log_size_bytes",
				Labels: append(labels, []string{"direction", "out"}),
				Value:  vFloat,
				Type:   "stats",
			})
		case "resp_time":
			m.Acc(&common.Metric{
				Name:   "log_duration_seconds",
				Labels: labels,
				Value:  float64(vFloat / logPattern.RespTimeDivider),
				Type:   "stats",
			})
		case "code":
			m.Acc(&common.Metric{
				Name:   "log_requests",
				Labels: labels,
				Value:  1,
				Type:   "counter",
			})
		}
	}

	return nil
}
